from django.contrib import admin

from .models import HelloSignRequest, HelloSignLog, HelloSignSigningUrl


class HelloSignRequestAdmin(admin.ModelAdmin):
    list_display = ('source_object', 'unclaimed_draft_guid', 'signature_request_id', 'dateof',)
    search_fields = ('signature_request_id',)


class HelloSignLogAdmin(admin.ModelAdmin):
    list_display = ('request', 'event_type', 'dateof',)
    list_filter = ['event_type']
    search_fields = ('event_type',)


class HelloSignSigningUrlAdmin(admin.ModelAdmin):
    list_display = ('request', 'signature_id', 'has_been_viewed', 'expires_at', 'dateof',)
    list_filter = ['has_been_viewed']
    search_fields = ('request', 'signature_id',)


admin.site.register(HelloSignRequest, HelloSignRequestAdmin)
admin.site.register(HelloSignLog, HelloSignLogAdmin)
admin.site.register(HelloSignSigningUrl, HelloSignSigningUrlAdmin)




# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.auth.models import User

from jsonfield import JSONField


class HelloSignRequest(models.Model):
    """
    Model to store the HelloSign Signing Request info
    """
    content_object_type = models.ForeignKey('contenttypes.ContentType', db_index=True)
    object_id = models.IntegerField(db_index=True)
    signature_request_id = models.CharField(null=True, blank=True, max_length=128, db_index=True)
    unclaimed_draft_guid = models.CharField(null=True, blank=True, max_length=128, db_index=True)
    dateof = models.DateTimeField(auto_now=False, auto_now_add=True, db_index=True)
    data = JSONField(default={})

    class Meta:
        # newest first
        ordering = ['-dateof']

    def __unicode__(self):
        return u'HelloSign Request for %s' % self.source_object

    @property
    def source_object(self):
        return self.content_object_type.get_object_for_this_type(pk=self.object_id)

    @property
    def is_claimed(self):
        return self.data.get('is_claimed', False)

    @property
    def claim_url(self):
        return self.data.get('claim_url', None)

    @property
    def signing_url(self):
        return self.data.get('signature_request', {}).get('signing_url', None)

    def get_absolute_url(self):
        url = None

        # has been claimed so we should have the data
        if self.is_claimed is True and self.data.get('signature_request', None) is not None:
            # return signing url
            url = self.signing_url

        elif self.unclaimed_draft_guid is not None and self.claim_url is not None:
            # return the unclaimed draft url
            url = self.claim_url

        return url


class HelloSignLog(models.Model):
    """
    Model to store the HelloSign Webhook Events
    """
    request = models.ForeignKey('hello_sign.HelloSignRequest')
    event_type = models.CharField(max_length=32, db_index=True)
    dateof = models.DateTimeField(auto_now=False, auto_now_add=True, db_index=True)
    data = JSONField(default={})

    class Meta:
        # newest first
        ordering = ['-id']

    def __unicode__(self):
        return u'%s (%s)' % (self.request, self.event_type)

    @property
    def response_data(self):
        return self.data['signature_request'].get('response_data', [])

    @property
    def signatures(self):
        return self.data['signature_request'].get('signatures', [])

    @property
    def signer_status(self):
        """
        Return the User object and status for this log
        """
        first = None
        response_data = self.response_data
        #
        # reversed is used because HS send the complete response_data
        # and simply keeps appending to it thus the most recent elements are
        # the last elements on the list
        #
        for item in reversed(response_data):
            #
            # get the item with the type "signature"
            # because all kinds of crazy are passed in from HS
            #
            if item.get('type') == "signature":
                first = item
                break

        #
        # we found a matching signature
        #
        if first is not None:
            signature_id = first.get('signature_id')

            for signer in self.signatures:
                signer_email_address = signer.get('signer_email_address')

                if signer.get('signature_id') == signature_id:
                    status_code = signer.get('status_code')
                    user = User.objects.get(email=signer_email_address)
                    return (user, status_code)

        return (None, None)


class HelloSignSigningUrl(models.Model):
    """
    Signing Url Record
    """
    request = models.ForeignKey('hello_sign.HelloSignRequest')
    signature_id = models.CharField(max_length=128, db_index=True)
    # set to False by the signals.on_signature_request_viewed_invalidate_signer_url
    has_been_viewed = models.BooleanField(default=False)
    expires_at = models.DateTimeField(auto_now=False, auto_now_add=False, null=True, db_index=True)
    dateof = models.DateTimeField(auto_now=False, auto_now_add=True, db_index=True)
    data = JSONField(default={})

    @property
    def sign_url(self):
        return self.data.get('sign_url')

    @property
    def has_expired(self):
        return self.expires_at < datetime.utcnow()


from .signals import on_signature_request_viewed_invalidate_signer_url



# -*- coding: utf-8 -*-
__version_info__ = ('0', '1', '5')
__version__ = '.'.join(__version_info__)

from django.conf import settings

import logging
logger = logging.getLogger('django.request')


HELLOSIGN_AUTHENTICATION = getattr(settings, 'HELLOSIGN_AUTHENTICATION', None)
HELLOSIGN_API_KEY = getattr(settings, 'HELLOSIGN_API_KEY', None)
HELLOSIGN_CLIENT_ID = getattr(settings, 'HELLOSIGN_CLIENT_ID', None)
HELLOSIGN_CLIENT_SECRET = getattr(settings, 'HELLOSIGN_CLIENT_SECRET', None)

assert HELLOSIGN_AUTHENTICATION is not None, 'You must define a settings.HELLOSIGN_AUTHENTICATION see: http://www.hellosign.com/api/reference#EventHashVerification *NB* ensure you are logged in as your primary account'
assert HELLOSIGN_API_KEY is not None, 'You must define a settings.HELLOSIGN_API_KEY see: http://www.hellosign.com/home/myAccount/current_tab/integrations'
assert HELLOSIGN_CLIENT_ID is not None, 'You must define a settings.HELLOSIGN_CLIENT_ID see: http://www.hellosign.com/home/myAccount/current_tab/integrations'
assert HELLOSIGN_CLIENT_SECRET is not None, 'You must define a settings.HELLOSIGN_CLIENT_SECRET see: http://www.hellosign.com/home/myAccount/current_tab/integrations'


# -*- coding: utf-8 -*-
from django.conf import settings
from django.contrib.contenttypes.models import ContentType

from jsonfield import JSONField

from hellosign import HelloSignUnclaimedDraftDocumentSignature
from .services import HelloSignService, HelloSignSignerService
from .models import HelloSignRequest

from . import logger

import json
import datetime

HELLOSIGN_TEST_MODE = getattr(settings, 'HELLOSIGN_TEST_MODE', 1)


class OverrideModelMethodException(Exception):
    message = 'You must override this method'


class ModelContentTypeMixin(object):
    """
    Mixin to provide get_content_type_object of the model its mixed in with
    """
    def get_content_type_object(self):
        """
        Required Method for acessing the current models ContentType
        """
        return ContentType.objects.get(app_label=self._meta.app_label,
                                       model=self._meta.model_name)


class HelloSignModelMixin(ModelContentTypeMixin):
    data = JSONField(default={})  # required to store data

    def hellosign_requests(self):
        """
        QuerySet of HelloSignRequest objects
        relies on HelloSignRequest._meta to define order
        """
        return HelloSignRequest.objects.filter(object_id=self.pk,
                                               content_object_type=self.get_content_type_object())

    @property
    def hellosign(self):
        """
        The Latest HelloSign Request
        """
        return self.hellosign_requests().first()

    @property
    def signing_data(self):
        return getattr(self.hellosign, 'data', None)

    @property
    def signature_request_id(self):
        return getattr(self.hellosign, 'signature_request_id', None)

    def sign_url_for_email(self, signer_email):
        """
        Get the sign_url for a specific email address
        """
        service = HelloSignSignerService(obj=self, signer_email=signer_email)
        return service.sign_url_for_signer(email=signer_email)

    @property
    def signatures(self):
        return self.signing_data.get('signatures', [])

    @signatures.setter
    def signatures(self, value):
        """
        Setter to allow us to update the signatures details segment of the
        current json object
        """
        obj = self.hellosign  # most recent HSRequest object
        obj.data['signatures'] = value
        obj.save(update_fields=['data'])

        # and the local model
        self.data['signatures'] = value
        self.save(update_fields=['data'])

    def hs_subject(self):
        """
        Simple subject sent as part of the HelloSign request used by HS
        in communication and in their own web interface
        """
        return 'Signature Request for %s' % self

    def hs_message(self):
        """
        Descriptive message sent as part of the HelloSign request used by HS
        in communication and in their own web interface
        """
        return 'Please review and sign this document at your earliest convenience'

    def hs_signers(self):
        """
        Return a list of invitees to sign
        Order here is important as it ties in tightly with the HS signers
        unfortunately
        """
        raise OverrideModelMethodException('You must override hs_signers and return a dict {"name": name, "email": email}')

    def hs_document_title(self):
        """
        Method to set the document title, displayed in the HelloSign Interface
        """
        raise OverrideModelMethodException('You must override hs_document_title and return the title of the document to send to HelloSign')

    def hs_document(self):
        """
        Return the document to be senf for signing
        """
        raise OverrideModelMethodException('You must override hs_document and return the pdf/docx/doc file to send to HelloSign')

    def get_hs_service(self, **kwargs):
        """
        Return the HelloSign service instance with all required to send for
        signing
        """
        kwargs.update({
            'document': self.hs_document(),
            'title': self.hs_document_title(),
            'invitees': self.hs_signers(),
            'subject': self.hs_subject(),
            'message': self.hs_message(),
        })
        return HelloSignService(**kwargs)

    def create_unclaimed_draft(self, requester_email_address, **kwargs):
        """
        Create an unsigned draft that can be used to setup a document for signing
        """
        service = self.get_hs_service(HelloSignSignatureClass=HelloSignUnclaimedDraftDocumentSignature)  # override the HelloSignSignatureClass

        resp = service.create_unclaimed_draft(test_mode=HELLOSIGN_TEST_MODE,
                                              client_id=settings.HELLOSIGN_CLIENT_ID,
                                              requester_email_address=requester_email_address,
                                              **kwargs)

        return self.hs_process_result(resp=resp)

    def send_for_signing(self, **kwargs):
        """
        Primary method used to send a document for signing
        """
        service = self.get_hs_service()

        resp = service.send_for_signing(test_mode=HELLOSIGN_TEST_MODE,
                                        client_id=settings.HELLOSIGN_CLIENT_ID,
                                        **kwargs)

        return self.hs_process_result(resp=resp)

    def hs_process_result(self, resp):
        if type(resp) not in [dict]:
            if hasattr(resp, 'json') is False:
                raise Exception('resp needs to be a dict or a requests response object with .json() method')
            else:
                resp = resp.json()  # assume is a requests response object

        logger.debug('HelloSign response: %s' % resp)

        # post process the result
        resp = self.hs_post_process_result(resp=resp)

        self.hs_record_result(result=resp)

        return resp

    def hs_post_process_result(self, resp):
        date_sent = str(datetime.datetime.utcnow())
        #
        # Add the date because HelloSign does not provide a date
        #
        resp.update({
            'date_sent': date_sent
        })

        return resp

    def hs_record_result(self, result):
        # setup the hs request object
        signature_request = result.get('signature_request', {})
        signature_request_id = signature_request.get('signature_request_id') # get id

        if signature_request_id:
            hs_request_object, is_new = HelloSignRequest.objects.get_or_create(signature_request_id=signature_request_id,
                                                                               content_object_type=self.get_content_type_object(),
                                                                               object_id=self.pk)
            #
            # Update the request object data dict with the signature_request object
            #
            data = hs_request_object.data
            data.update(result)
            # save it back
            hs_request_object.data = data
            hs_request_object.save(update_fields=['data'])

            return hs_request_object
        return None



# -*- coding: utf-8 -*-
from django.conf import settings
from django.utils.timezone import utc
from hellosign import HelloSigner, HelloDoc
from hellosign import (HelloSignSignature,
                       HelloSignEmbeddedDocumentSignature,
                       HelloSignEmbeddedDocumentSigningUrl,
                       HelloSignUnclaimedDraftDocumentSignature)

import time
from datetime import timedelta, datetime

from . import logger
from .models import HelloSignSigningUrl


class AuthenticationSettingsException(Exception):
    message = 'settings.HELLOSIGN_AUTHENTICATION tuple ("username", "password") has not been provided.'


class BaseHelloSignHelper(object):
    hellosign_authentication = None

    def __init__(self, **kwargs):
        try:
            self.hellosign_authentication = settings.HELLOSIGN_AUTHENTICATION
            assert type(self.hellosign_authentication) == tuple
            assert len(self.hellosign_authentication) == 2
        except AttributeError:
            msg = 'settings.HELLOSIGN_AUTHENTICATION tuple ("username", "password") has not been provided.'
            logger.critical(msg)
            raise AuthenticationSettingsException(msg)

    def validate(self, resp):
        if 'error' in resp or resp.status_code not in [200, 201, 202]:
            raise Exception('HelloSign Api Error: %s' % resp)

        return resp.json()


class HelloSignService(BaseHelloSignHelper):
    """
    Service that allows us to send a document for signing
    """
    def __init__(self, document, invitees, **kwargs):
        self.document = document
        logger.info(u'Submitting document to HelloSign: "%s"' % document.name)

        # Dependency injected class for testing and using other HS Classes
        self.HelloSignSignatureClass = kwargs.get('HelloSignSignatureClass', HelloSignEmbeddedDocumentSignature)

        self.invitees = invitees
        self.title = kwargs.get('title', self.document.name)
        self.user = kwargs.get('user', None)
        self.subject = kwargs.get('subject', None)
        self.message = kwargs.get('message', None)

        super(HelloSignService, self).__init__(**kwargs)

    def create_unclaimed_draft(self, **kwargs):
        unclaimed_draft = HelloSignUnclaimedDraftDocumentSignature(title=self.title, subject=self.subject, message=self.message)

        # Add invitees
        for i in self.invitees:
            unclaimed_draft.add_signer(HelloSigner(name=i['name'], email=i['email']))

        unclaimed_draft.add_doc(HelloDoc(file_path=self.document.name))

        # Perform the submission
        data = {}

        resp = self.validate(unclaimed_draft.create(auth=self.hellosign_authentication, **kwargs))
        data.update(resp)

        # extract the signature_request_id
        signature_request_id = data.get('unclaimed_draft', {}).get('signature_request_id', None)

        return data

    def send_for_signing(self, **kwargs):
        signature_request_id = kwargs.pop('signature_request_id', None)

        if signature_request_id is not None:
            #
            # Existing signing request use details
            #
            signature = self.HelloSignSignatureClass(signature_request_id=signature_request_id)
            # get the detail
            return signature.detail(signature_request_id=signature_request_id,
                                    auth=self.hellosign_authentication)
        else:
            #
            # Is a new signature request add signers and add the doc and 
            # use create
            signature = self.HelloSignSignatureClass(title=self.title, subject=self.subject, message=self.message)
            #
            # Add invitees
            for i in self.invitees:
                signature.add_signer(HelloSigner(name=i['name'], email=i['email']))

            signature.add_doc(HelloDoc(file_path=self.document.name))

            # Perform the submission
            return self.validate(signature.create(auth=self.hellosign_authentication, **kwargs))


class HelloSignSignerService(BaseHelloSignHelper):
    """
    Service that returns various data for and about the signer
    """
    signer_email = None  # the current users email
    obj = None
    hellosign_request = None
    signatures = None
    signatures = []  # the current set of signatures for a request

    def __init__(self, obj, signer_email=None, **kwargs):
        self.signer_email = signer_email
        self.obj = obj
        self.hellosign_request = obj.hellosign # method on mixin to get latest HS response object
        self.signatures = obj.signatures[::]

        super(HelloSignSignerService, self).__init__(**kwargs)

        self.process()

    def embedded_signature_url(self, signature_id):
        service = HelloSignEmbeddedDocumentSigningUrl(signature_id=signature_id)
        return self.validate(service.create(auth=self.hellosign_authentication))

    def process(self):
        """
        Get the embedded signature url and expiry from HS
        @TODO refactor @CODESMELL
        """
        for i, signer in enumerate(self.signatures):
            logger.debug(u'HelloSign get a signing_url for signer: %s' % signer)
            #
            # If we have NO email address specified then update all the signer dicts
            # if we have an email and the current signers email matches it then update
            # just that one
            #
            if self.signer_email is None or self.signer_email == signer.get('signer_email_address'):

                status_code = signer.get('status_code', None)
                expires_at = signer.get('expires_at', None)
                logger.debug('Get signing_url for status_code: %s, expires_at: %s' % (status_code, signer))

                #
                # If we are waiting on a signature or have no code
                #
                if status_code in ['awaiting_signature', None]:
                    logger.debug(u'signing_url status_code is valid and we can try get a signature')
                    #
                    # If we have no expries_at date (implies url has already been got)
                    # or if the expires at has expired
                    #
                    if expires_at is None or datetime.fromtimestamp(int(expires_at)) >= datetime.utcnow():
                        logger.debug(u'can get signature url, as current is None or has expired')

                        resp = self.embedded_signature_url(signature_id=signer.get('signature_id'))
                        signing_url_log = self.process_embedded_signature_url(signature=signer, resp=resp)
                        #
                        # These are critical 
                        # the sign_url gets used in the search for the users sign_url
                        # in self.sign_url_for_signer
                        #
                        signer['sign_url'] = signing_url_log.sign_url
                        signer['expires_at'] = signing_url_log.expires_at

                    else:
                        logger.info('signing_url status_code is invalid has not expired: %s for : %s' % (expires_at, self.signatures[i]))
                        pass

    def process_embedded_signature_url(self, signature, resp):
        """
        Process the HS response to an embeded signature url
        """
        resp_json = resp.get('embedded', {})

        signature_id = signature.get('signature_id')

        expires_at = datetime.fromtimestamp(resp_json.get('expires_at')).replace(tzinfo=utc)
        #
        # Create a SignUrl Object for the record
        #
        signing_url_log, is_new = HelloSignSigningUrl.objects.get_or_create(request=self.hellosign_request,
                                                                            signature_id=signature_id)
        signing_url_log.expires_at = expires_at
        signing_url_log.data = resp_json
        signing_url_log.save(update_fields=['expires_at', 'data'])

        logger.debug('Record the sign_url object: %s, is_new: %s' % (signing_url_log, is_new))
        return signing_url_log

    def sign_url_for_signer(self, email):
        """
        Loop over the provided signatures and look for the specific email
        """
        logger.debug(u'Finding sign_url_for_signer: %s' % email)

        for i, signer in enumerate(self.signatures):
            status_code = signer.get('status_code', None)
            logger.debug(u'Status code for sign_url_for_signer: %s' % status_code)

            if status_code in ['awaiting_signature', None]:
                logger.debug(u'Valid status code for sign_url_for_signer: %s' % status_code)

                if signer.get('signer_email_address') == email:

                    sign_url = signer.get('sign_url')
                    logger.debug('sign_url for: %s is %s' % (email, sign_url))

                    return sign_url
        return None




# -*- coding: utf-8 -*-
"""
HelloSign webhook Event
"""
from django.dispatch import Signal, receiver

hellosign_webhook_event_recieved = Signal(providing_args=['signature_request_id',
                                                          'event_type',
                                                          'data',
                                                          'hellosign_request',
                                                          'hellosign_log'])


@receiver(hellosign_webhook_event_recieved)
def on_signature_request_viewed_invalidate_signer_url(sender, hellosign_log, signature_request_id, hellosign_request, event_type, data, **kwargs):
    """
    Invalidate the signer url when we recieve  signature_request_viewed event
    """
    if event_type == 'signature_request_viewed':
        #
        # should be able to delete the HelloSignSigningUrl object for this
        # event; but there is nothing to bund the HS event object
        # with our HelloSignSigningUrl.signature_id and
        # HelloSignSigningUrl.request object
        #
        pass



# -*- coding: utf-8 -*-
import os
from setuptools import setup

setup(
    name = "django-hello_sign",
    version = "0.2.0",
    author = "Ross Crawford-d'Heureuse",
    author_email = "ross@lawpal.com",
    description = ("Django app for integrating with HelloSign"),
    license = "MIT",
    keywords = "django hellosign app",
    url = "https://github.com/rosscdh/django-hello_sign",
    packages=['hello_sign'],
    install_requires = [
        'hellosign',  # must refer to package version explicitly **required**
    ]
)



# -*- coding: utf-8 -*-
from django.conf.urls import patterns, url, include
from django.views.decorators.csrf import csrf_exempt

from .views import HelloSignWebhookEventHandler


urlpatterns = patterns('',
    url(r'^event/$', csrf_exempt(HelloSignWebhookEventHandler.as_view()), name='hellosign_webhook_event'),
)



# -*- coding: utf-8 -*-
from django.conf import settings
from django.views.generic import CreateView
from django.http import HttpResponseBadRequest
from django.shortcuts import render, get_object_or_404

import hashlib, hmac

from . import logger

from .models import HelloSignRequest, HelloSignLog
from .signals import hellosign_webhook_event_recieved

import json
import logging
logger = logging.getLogger('django.request')


class HelloSignWebhookEventHandler(CreateView):
    """
    View to handle the HelloSign webhook events

    as per: http://www.hellosign.com/api/gettingStarted "Handling callbacks"

    NOTE: Your endpoint will need to return a 200 HTTP code and a response body
    containing the following text: Hello API Event Received. Otherwise,
    the callback will be considered a failure and will be retried later.
    """
    model = HelloSignLog
    # response needs to return "Hello API Event Received"
    template_name = 'sign/hello_sign/webhook_create_event.html'

    def validate_callback(self, event_data):
        # expected result
        event_hash = event_data['event'].get('event_hash')
        # comparison
        event_type = event_data['event'].get('event_type')
        event_time = event_data['event'].get('event_time')

        assert event_hash == unicode(hmac.new(settings.HELLOSIGN_API_KEY, (event_time + event_type), hashlib.sha256).hexdigest()), 'event_hash does not match see: https://www.hellosign.com/api/eventsAndCallbacksWalkthrough#EventHash'
        # if event_hash != unicode(hmac.new(settings.HELLOSIGN_API_KEY, (event_time + event_type), hashlib.sha256).hexdigest()):
        #     raise Exception('event_hash does not match see: https://www.hellosign.com/api/eventsAndCallbacksWalkthrough#EventHash')


    def extract_json_data(self, body):
        logger.debug(u'Post from HelloSign: %s' % body)

        try:
            data = json.loads(body)

            if 'event' not in data:
                raise Exception('No event key found in HelloSign Post body')

            if 'signature_request' not in data:
                logger.info('No signature_request key found in HelloSign Post body')

        except Exception as e:
            logger.critical(u'Could not extract json from request.body in HelloSignWebhookEventHandler: %s' % e)
            raise e

        return data

    def post(self, request, *args, **kwargs):
        """
        NB: HelloSign sends a post object with key "json" that is set to an actual
        string of JSON
        """
        data = self.extract_json_data(body=request.POST.get('json'))  # extract json

        signature_request_id = data.get('signature_request', {}).get('signature_request_id', None)

        event_type = data['event'].get('event_type')

        if signature_request_id is None:
            logger.error('Hellosign signature_request_id: %s is None' % signature_request_id)

        else:
            # validate callback
            try:
                self.validate_callback(event_data=data)

            except Exception as e:
                return HttpResponseBadRequest('HelloSign webhook exception: %s' % e)

            logger.info('recieved webhook event: %s from HelloSign signature_request_id: %s' % (event_type, signature_request_id))

            # get the request object which must exist as its created when the object is sent for signing
            hellosign_request = get_object_or_404(HelloSignRequest, signature_request_id=signature_request_id)

            # create log object
            self.object = self.model.objects.create(request=hellosign_request,
                                                    event_type=event_type,
                                                    data=data)

            logger.info(u'Issuing hellosign_webhook_event_recieved signal')
            hellosign_webhook_event_recieved.send(sender=self,
                                                  signature_request_id=signature_request_id,
                                                  event_type=event_type,
                                                  data=data,
                                                  hellosign_request=hellosign_request,
                                                  hellosign_log=self.object)

        return render(request, self.template_name)
